
(defun parse-observations (OBSERVATION-LIST)
  (declare (special ObservedPredicates ObservedLiterals))
  (cond ((null OBSERVATION-LIST) nil)
        (t (setf (gethash (caar OBSERVATION-LIST) ObservedPredicates) 1)
           (setf (gethash (car OBSERVATION-LIST) ObservedLiterals) 1)
           (parse-observations (cdr OBSERVATION-LIST)))))

;
; Recursive version blew up recursion stack
;
; (defun remove-valid-clauses (CL)
;  (cond ((null CL) nil)
;	((valid (car CL)) (remove-valid-clauses (cdr CL)))
;	(t (cons (car CL) (remove-valid-clauses (cdr CL))))))
;
(defun remove-valid-clauses (CL)
  (let (answer)
    (dolist (c CL)
      (if (null (valid c))
	  (setq answer (cons c answer))))
    answer))

(defun valid (C)
  (cond ((null C) nil)
	((member (complement-literal (car C)) (cdr C) :test #'equal) t)
	(t (valid (cdr C)))))

(defun complement-literal (L)
  (cond ((atom L) (list 'not L))
	((eql (car L) 'not) (cadr L))
	(t (list 'not L))))

(defun parse-schema-list (SCHEMA-LIST)
  (cond ((null SCHEMA-LIST) nil)
        (t (append (parse-schema (car SCHEMA-LIST)) 
		   (parse-schema-list (cdr SCHEMA-LIST))))))

(defun parse-schema (SCHEMA)
  (cond ((atom SCHEMA) (parse-formula SCHEMA))
	((eql (car SCHEMA) 'type) (parse-type (cdr SCHEMA)))
	((eql (car SCHEMA) 'option) (parse-option (cdr SCHEMA)))
        (t (parse-formula SCHEMA))))

(defun parse-option (ARGS)
  (let ((opt (car ARGS))
	(val (parse-expression (cadr ARGS))))
    (if (eql val 0) (setq val nil))
    (cond ((member opt '(compact-encoding))
	   (set opt val))
	  (t (error "Cannot parse option ~S" ARGS)))
    nil))

(defun parse-type (DEFINITION)
  (bind-uniquely (car DEFINITION)
                 (parse-expression (cadr DEFINITION)) t)
  nil)

(defun bind-uniquely (VAR VAL &optional noerror)
  (declare (special Bind))
  (if (gethash VAR Bind) 
      (if noerror
          nil
	(error "Cannot shadow variable ~S with binding ~S" VAR VAL)) 
    (setf (gethash VAR Bind) VAL)))

(defun unbind (VAR)
  (declare (special Bind))
  (remhash VAR Bind))

(defun is-bound (VAR)
  (declare (special Bind))
  (gethash VAR Bind))

(defun binding-of (VAR)
  (declare (special Bind))
  (gethash VAR Bind))

(defun is-observed-literal (F)
  (declare (special ObservedPredicates))
  (and (is-literal F)
       (listp F)
       (if (eql (car F) 'not)
           (gethash (caadr F) ObservedPredicates)
         (gethash (car F) ObservedPredicates))))

(defun parse-observed-literal (F)
  (declare (special ObservedLiterals))
  (let ((answ (cond ((eql (car F) 'not) 
                     (if (is-true (gethash (parse-literal (cadr F)) ObservedLiterals 0))
                         '(())
                       '()))
                    (t
                     (if (is-true (gethash (parse-literal F) ObservedLiterals 0))
                         '()
                       '(()))))))
    answ))

(defun parse-formula (F)
  ;; (format t "entering parse ~S" F)
  (cond ((is-observed-literal F) (parse-observed-literal F))
        ((is-literal F) (list (list (parse-literal F))))
	((eql (car F) 'not) (parse-not (cadr F)))
	((eql (car F) 'and) (parse-and (cdr F)))
	((eql (car F) 'or) (parse-or (cdr F)))
	((eql (car F) 'implies) (parse-implies (cdr F)))
	((eql (car F) 'if) (parse-if (cadr F) (caddr F)))
	((eql (car F) 'equiv) (parse-equiv (cdr F)))
	((eql (car F) 'all) 
	 (parse-forall (cadr F) 
		       (parse-set-expression (caddr F))
		       (cadddr F) 
		       (car (cddddr F))))
	((eql (car F) 'exists) 
	 (parse-forsome (cadr F) 
		       (parse-set-expression (caddr F))
		       (cadddr F) 
		       (car (cddddr F))))
	(t (error "Cannot parse formula ~S" F))))

(defun parse-if (test body)
  (cond ((is-false (parse-integer-expression test)) nil)
	(t (parse-formula body))))

(defun parse-not (F)
  ;; F is not a literal, that case is handled in parse-formula
  ;; (format t "entering parse-not ~S" F)
  (let ((op (car F)))
    (cond ((eql op 'not) (parse-formula (cadr F)))
          ((eql op 'and) (parse-formula (cons 'or (negate-list (cdr F)))))
          ((eql op 'or) (parse-formula (cons 'and (negate-list (cdr F)))))
          ((eql op 'implies) (parse-formula (list 'and (cadr F) (list 'not (caddr F)))))
	  ((eql op 'if) (cond ((is-true (parse-integer-expression (cadr F))) (parse-formula `(not ,(caddr F))))
			      (t '(()))))
          ((eql op 'equiv) (append (parse-formula `(or ,(cadr F) ,(caddr F)))
                                   (parse-formula `(or (not ,(cadr F)) (not ,(caddr F))))))
          ((eql op 'all) (parse-formula `(exists ,(cadr F) ,(caddr F) ,(cadddr F) (not ,(car (cddddr F))))))
          ((eql op 'exists) (parse-formula `(all ,(cadr F) ,(caddr F) ,(cadddr F) (not ,(car (cddddr F))))))
          (t (error "Cannot parse negation ~S" F)))))

(defun negate-list (L)
     (cond ((null L) nil)
           (t (cons (list 'not (car L)) (negate-list (cdr L))))))

(defun parse-implies (FL)
  (if (not (= (length FL) 2)) (error "Cannot parse implication ~S" FL))
  (parse-or (cons (list 'not (car FL)) (cdr FL))))

(defun parse-equiv (FL)
  (if (not (= (length FL) 2)) (error "Cannot parse equivalence ~S" FL))
  (append (parse-implies FL)
          (parse-formula `(implies ,(cadr FL) ,(car FL)))))

(defun is-literal (F) 
  (or (is-proposition F)
      (and (eql 'not (car F)) (is-proposition (cadr F)))))

(defun is-proposition (F)
  (or (atom F)
      (not (member (car F) logical-connectives))))

(defun parse-and (FL) ;; and just appends the clauses
  (parse-schema-list FL))

(defun parse-or (FL)
  (cond ((null FL) (list nil)) ;; empty OR is the empty clause
	(t (multiply-clauses (parse-schema (car FL)) 
			     (parse-or (cdr FL))))))

(defun multiply-clauses (L R)
  (if (or (null compact-encoding)
	  (< (length L) 2)
	  (< (length R) 2)
	  (< (+ (length L) (length R)) 5))
      (explicit-multiply-clauses L R)
      (let ((g (gensym "XX"))) ;; g selects whether L or R must be true
	(append (mapcar #'(lambda (c) (cons g c)) R) 
		(mapcar #'(lambda (c) (cons (list 'not g) c)) L)))))

(defun merge-clauses (C1 C2)
  (remove-duplicates (append C1 C2) :test #'equal))

; Reduce stack requirements - rewrite following two recursive functions
;  as a single iterative function.
;
;(defun explicit-multiply-clauses (L R)
;  (cond ((null L) nil)
;	(t (append (multiply-one-clause (car L) R)
;		   (explicit-multiply-clauses (cdr L) R)))))
;
;(defun multiply-one-clause (C R)
;  (cond ((null R) nil)
;	(t (append (list (merge-clauses C (car R)))
;		   (multiply-one-clause C (cdr R))))))


(defun explicit-multiply-clauses (L R)
  (let (answer)
    (dolist (lclause L)
      (setq answer (cons (multiply-one-clause lclause R) answer)))
    (mapcan #'copy-list answer)))

(defun multiply-one-clause (C R)
  (let (answer)
    (dolist (rclause R)
      (setq answer (cons (merge-clauses C rclause) answer)))
    answer))

(defun parse-forall (VAR DOM TEST BODY)
  (cond ((null DOM) nil) ;; the empty list of clauses
        (t (append (parse-binding VAR (car DOM) TEST BODY nil)
                   (parse-forall VAR (cdr DOM) TEST BODY)))))

(defun is-false (x)
  (or (null x) (eql x 0)))

(defun is-true (x)
  (not (is-false x)))

(defun parse-binding (VAR VAL TEST BODY FAILED-TEST-RESULT)
  (declare (special Bind))
  (let ((RESULT FAILED-TEST-RESULT))
    (bind-uniquely VAR VAL)
    (let ((TESTVAL (parse-expression TEST)))
      (if (is-true TESTVAL)
          (setq RESULT (parse-schema BODY))))
    (unbind VAR)
    RESULT))

(defun parse-forsome (VAR DOM TEST BODY)
  (cond ((NULL Dom) (list nil)) ;; the empty clause
        (t (multiply-clauses (parse-binding VAR (car DOM) TEST BODY (list nil))
			     (parse-forsome VAR (cdr DOM) TEST BODY)))))

(defun parse-set-expression (EXPR)
  (let ((answ (parse-expression EXPR)))
    (if (not (listp answ)) (error "Set expected instead of ~S" EXPR))
    answ))

(defun parse-integer-expression (EXPR)
  (let ((answ (parse-expression EXPR)))
    (if (not (integerp answ)) (error "Integer expected instead of ~S" EXPR))
    answ))

(defun parse-name (EXPR)
  (if (or (null EXPR)
	  (integerp EXPR)
	  (listp EXPR)
	  (member EXPR reserved-words))
      (error "Symbol expected instead of ~S" EXPR))
    answ))

(defun parse-or-expression (EXPR)
  (cond ((null EXPR) 0)
	((is-true (parse-expression (car EXPR))) 1)
	(t (parse-or-expression (cdr EXPR)))))

(defun parse-and-expression (EXPR)
  (cond ((null EXPR) 1)
	((is-false (parse-expression (car EXPR))) 0)
	(t (parse-and-expression (cdr EXPR)))))

(defun evaluate-lisp-expression (EXPR)
  (declare (special Bind))
  (maphash #'set Bind)
  (eval EXPR))

(defun parse-expression (EXPR)
  (declare (special Bind ObservedPredicates))
  (cond ((and (symbolp EXPR) (not (null EXPR)) (is-bound EXPR)) (binding-of EXPR))
        ((atom EXPR) EXPR)
        (t (let ((op (car EXPR)))
             (cond ((eql op 'not) (if (is-true (parse-expression (cadr EXPR))) 0 1))
		   ((eql op 'and) (parse-and-expression (cdr EXPR)))
		   ((eql op 'or) (parse-or-expression (cdr EXPR)))
                   ((eql op 'set) (parse-enumerated-set (cdr EXPR)))	
                   ((gethash op ObservedPredicates)
                    (parse-observed-literal-expression EXP))
		   ((eql op 'lisp)
		    (evaluate-lisp-expression (cadr EXPR)))
		   ((and (member op binary-operators) (= (length EXPR 3)))
		    (parse-binary-expression op (cadr EXPR) (caddr EXPR)))
		   (t (error "Parser error at ~S" EXPR)))))))
