;;; Example:
;;; (parse '((definition boys (set sam alex jon)) (forall x boys t (equiv (lonely x) (not (forsome y boys (neq x y) (friend x y)))))) nil)
;;; ((OR (NOT (LONELY SAM)) (NOT (FRIEND SAM ALEX)))
;;;  (OR (NOT (LONELY SAM)) (NOT (FRIEND SAM JON)))
;;;  (OR (FRIEND SAM ALEX) (FRIEND SAM JON) (LONELY SAM))
;;;  (OR (NOT (LONELY ALEX)) (NOT (FRIEND ALEX SAM)))
;;;  (OR (NOT (LONELY ALEX)) (NOT (FRIEND ALEX JON)))
;;;  (OR (FRIEND ALEX SAM) (FRIEND ALEX JON) (LONELY ALEX))
;;;  (OR (NOT (LONELY JON)) (NOT (FRIEND JON SAM)))
;;;  (OR (NOT (LONELY JON)) (NOT (FRIEND JON ALEX)))
;;;  (OR (FRIEND JON SAM) (FRIEND JON ALEX) (LONELY JON)))
;;; 
;;; (parse '((forall x (range 1 5) t (forall y (range 1 5) t (if (neq x y) (different x y))))) nil)
;;; ((OR (DIFFERENT 1 2)) (OR (DIFFERENT 1 3)) (OR (DIFFERENT 1 4)) (OR (DIFFERENT 1 5))
;;;  (OR (DIFFERENT 2 1)) (OR (DIFFERENT 2 3)) (OR (DIFFERENT 2 4)) (OR (DIFFERENT 2 5))
;;;  (OR (DIFFERENT 3 1)) (OR (DIFFERENT 3 2)) (OR (DIFFERENT 3 4)) (OR (DIFFERENT 3 5))
;;;  (OR (DIFFERENT 4 1)) (OR (DIFFERENT 4 2)) (OR (DIFFERENT 4 3)) (OR (DIFFERENT 4 5))
;;;  (OR (DIFFERENT 5 1)) (OR (DIFFERENT 5 2)) (OR (DIFFERENT 5 3))
;;;  (OR (DIFFERENT 5 4)))
;;; 
;;; (parse '((forall b (set sam alex) t (forall g (set mary sue) (likes b g) (dates b g)))) '((likes sam sue) (likes alex mary) (likes alex sue)))
;;;    ((OR (DATES SAM SUE)) (OR (DATES ALEX MARY)) (OR (DATES ALEX SUE)))  
;;;
;;; (parse '((forall b (set sam alex) t (forall g (set mary sue) (or (likes b g) (likes g b)) (dates b g)))) '((likes sam sue) (likes alex mary) (likes sue sam)))
;;; (parse '((forall b (set sam alex) t (forall g (set mary sue) (and (likes b g) (likes g b)) (dates b g)))) '((likes sam sue) (likes alex mary) (likes sue sam)))
;;;
;;;
;;; (parse '((or p (likes alex mary))) '((likes alex mary)))
;;; (parse '((or p (likes alex mary) q)) '((likes alex sue)))
;;; (parse '((or p (not (likes alex mary)))) '((likes alex sue)))

;;; Test of observed predicates in clauses:
;;; (parse '((or q (likes alex mary)) (or r (not (likes alex mary))) (or s (likes sam mary))) '((likes sam sue) (likes alex mary) (likes sue sam)))
;;;
;;;   26 Feb 2010
;;;      - fixed bug for expression (range ...)
;;;      - added observed predicates
;;;   2 Mar 2010
;;;      - evaluates observed predicates when they appear in the body of a clause
;;;   8 Apr 2011
;;;      - replaced assoc list with hash table

(defvar compact-encoding nil) 

(defun propositionalize (WFFFILE CNFFILE MAPFILE)
  (with-open-file (WS WFFFILE :direction :input)
    (with-open-file (CS CNFFILE :direction :output :if-exists :supersede)
       (with-open-file (MS MAPFILE :direction :output :if-exists :supersede)
          (multiple-value-bind (cnfdata mapdata numvar numclauses)
              (lit2prop (loop with clause while (not (eql :EOF (setq clause (read WS nil :EOF))))
                              collect clause))
            (format CS "p cnf ~S ~S~%" numvar numclauses)
            (loop for c in cnfdata do (format CS "~{~D ~}0~%" c))
            (format MS "map ~S~%" numvar)
            (loop for m in mapdata do (format MS "~{~D ~S~}~%" m)))))))

(defun lit2prop (CL)
  (let ((cnfdata nil) (mapdata nil) (numvar 0) (numclauses (length CL)) (hash (make-hash-table :test #'equal)))
    ;; Build hash table
    (loop for clause in CL do
          (loop for lit in (cdr clause) do
                (let ((prop (if (is-proposition lit) lit (cadr lit))))
		  (cond ((not (nth-value 1 (gethash prop hash)))
			 (incf numvar)
			 (setf (gethash prop hash) numvar))))))
    ;; Translate clauses
    (setq cnfdata (loop for clause in CL 
                        collect
                        (loop for lit in (cdr clause) 
                              collect (if (is-proposition lit) 
                                          (gethash lit hash)
                                        (- (gethash (cadr lit) hash))))))
    ;; Build map table
    (maphash #'(lambda (key val) (push (list val key) mapdata)) hash)
    (setq mapdata (sort mapdata #'< :key #'car))
    ;; Return multiple values
    (values cnfdata mapdata numvar numclauses)))

(defun soln2lit (mapdata solndata)
  ;; return a list of propositions
  (let ((hash (make-hash-table)) proplist)
    (loop for pair in mapdata do (setf (gethash (car pair) hash) (cadr pair)))
    (setq proplist (loop for i in solndata collect (gethash i hash)))
    (setq proplist (sort proplist #'prop-ordering))
    proplist))

(defun prop-ordering (p q)
  (prop-order-r (if (atom p) p (reverse p))
                (if (atom q) q (reverse q))))

(defun prop-order-r (p q)
  (cond ((and (integerp p) (integerp q)) (< p q))
        ((and (atom p) (atom q)) (string-lessp p q))
        ((atom p) t)
        ((atom q) nil)
        ((equal (car p) (car q)) (prop-order-r (cdr p) (cdr q)))
        (t (prop-order-r (car p) (car q)))))
        

(defun interpret (SOLNFILE MAPFILE OUTFILE)
  (let (solndata mapdata litdata)
    ;; Read solnfile to create solution list.  Ignore any non-integers and negative integers in solnfile.
    (with-open-file 
     (SS SOLNFILE :direction :input)
     (setq solndata (loop with s while (not (eql :EOF (setq s (read SS nil :EOF)))) collect s))
     (setq solndata (remove-if-not #'(lambda (x) (and (integerp x) (> x 0))) solndata)))
    ;; Read mapfile to create mapdata list
    (with-open-file 
     (MS MAPFILE :direction :input)
     (if (not (eql (read MS) 'map)) (error "Bad map file"))
     (if (not (integerp (read MS))) (error "Bad map file"))
     (setq mapdata (loop with i with p 
                         while (not (eql :EOF (setq i (read MS nil :EOF))))
                         do (setq p (read MS)) (if (not (integerp i)) (error "Bad map file"))
                         collect (list i p))))
    ;; call soln2lit to create list of true literals, sorted by last argument and predicate name
    (setq litdata (soln2lit mapdata solndata))
    ;; Print list of true literals to outfile
    (with-open-file 
     (OS OUTFILE :direction :output :if-exists :supersede)
     (format OS "~{~S~%~}" litdata))
  ))


(defun instantiate (INFILE OUTFILE &optional OBSFILE)
  (with-open-file (INS INFILE :direction :input)
      (with-open-stream (OBS (if OBSFILE (open OBSFILE :direction :input) (make-concatenated-stream)))
           (with-open-file (OUTS OUTFILE :direction :output :if-exists :supersede)
               (let (CL SCHEMA OBSERVATION)
                 (setq CL (parse 
                           (loop while (not (eql 'EOF (setq SCHEMA (read INS nil 'EOF))))
			      collect SCHEMA)
			   (loop while (not (eql 'EOF (setq OBSERVATION (read OBS nil 'EOF))))
                                 collect OBSERVATION)))
                 (loop for C in CL do (format OUTS "~S~%" C)))))))

(defun parse (SCHEMA-LIST OBSERVATION-LIST)
  (let ((Bind (make-hash-table :test #'eql))
        (ObservedPredicates (make-hash-table :test #'eql))
        (ObservedLiterals (make-hash-table :test #'equal)))
    (declare (special Bind ObservedPredicates ObservedLiterals))
    (parse-observations OBSERVATION-LIST)
    (mapcar #'(lambda (c) (cons 'or c))
	    (remove-valid-clauses
	     (parse-schema-list SCHEMA-LIST)))))

(defun parse-observations (OBSERVATION-LIST)
  (declare (special ObservedPredicates ObservedLiterals))
  (cond ((null OBSERVATION-LIST) nil)
        (t (setf (gethash (caar OBSERVATION-LIST) ObservedPredicates) 1)
           (setf (gethash (car OBSERVATION-LIST) ObservedLiterals) 1)
           (parse-observations (cdr OBSERVATION-LIST)))))

; killer?
;(defun remove-valid-clauses (CL)
;  (cond ((null CL) nil)
;	((valid (car CL)) (remove-valid-clauses (cdr CL)))
;	(t (cons (car CL) (remove-valid-clauses (cdr CL))))))

(defun remove-valid-clauses (CL)
       CL)

(defun valid (C)
  (cond ((null C) nil)
	((member (complement-literal (car C)) (cdr C) :test #'equal) t)
	(t (valid (cdr C)))))

(defun complement-literal (L)
  (cond ((atom L) (list 'not L))
	((eql (car L) 'not) (cadr L))
	(t (list 'not L))))

(defun parse-schema-list (SCHEMA-LIST)
  (cond ((null SCHEMA-LIST) nil)
        (t (append (parse-schema (car SCHEMA-LIST)) 
		   (parse-schema-list (cdr SCHEMA-LIST))))))

(defun parse-schema (SCHEMA)
  (cond ((atom SCHEMA) (parse-formula SCHEMA))
	((eql (car SCHEMA) 'definition) (parse-definition (cdr SCHEMA)))
	((eql (car SCHEMA) 'option) (parse-option (cdr SCHEMA)))
        (t (parse-formula SCHEMA))))

(defun parse-option (ARGS)
  (let ((opt (car ARGS))
	(val (parse-expression (cadr ARGS))))
    (if (eql val 0) (setq val nil))
    (cond ((member opt '(compact-encoding))
	   (set opt val))
	  (t (error "Cannot parse option ~S" ARGS)))
    nil))

(defun parse-definition (DEFINITION)
  (bind-uniquely (car DEFINITION)
                 (parse-expression (cadr DEFINITION)))
  nil)


(defun bind-uniquely (VAR VAL)
  (declare (special Bind))
  (if (gethash VAR Bind) (error "Cannot shadow variable ~S with binding ~S" VAR VAL))
  (setf (gethash VAR Bind) VAL))

(defun unbind (VAR)
  (declare (special Bind))
  (remhash VAR Bind))

(defun is-bound (VAR)
  (declare (special Bind))
  (gethash VAR Bind))

(defun binding-of (VAR)
  (declare (special Bind))
  (gethash VAR Bind))


(defun is-observed-literal (F)
  (declare (special ObservedPredicates))
  (and (is-literal F)
       (listp F)
       (if (eql (car F) 'not)
           (gethash (caadr F) ObservedPredicates)
         (gethash (car F) ObservedPredicates))))

(defun parse-observed-literal (F)
  (declare (special ObservedLiterals))
  (let ((answ (cond ((eql (car F) 'not) 
                     (if (is-true (gethash (parse-literal (cadr F)) ObservedLiterals 0))
                         '(())
                       '()))
                    (t
                     (if (is-true (gethash (parse-literal F) ObservedLiterals 0))
                         '()
                       '(()))))))
    answ))


(defun parse-formula (F)
  ;; (format t "entering parse ~S" F)
  (cond ((is-observed-literal F) (parse-observed-literal F))
        ((is-literal F) (list (list (parse-literal F))))
	((eql (car F) 'not) (parse-not (cadr F)))
	((eql (car F) 'and) (parse-and (cdr F)))
	((eql (car F) 'or) (parse-or (cdr F)))
	((eql (car F) 'implies) (parse-implies (cdr F)))
	((eql (car F) 'if) (parse-if (cadr F) (caddr F)))
	((eql (car F) 'equiv) (parse-equiv (cdr F)))
	((eql (car F) 'forall) 
	 (parse-forall (cadr F) 
		       (parse-set-expression (caddr F))
		       (cadddr F) 
		       (car (cddddr F))))
	((eql (car F) 'forsome) 
	 (parse-forsome (cadr F) 
		       (parse-set-expression (caddr F))
		       (cadddr F) 
		       (car (cddddr F))))
	(t (error "Cannot parse formula ~S" F))))

(defun parse-if (test body)
  (cond ((is-false (parse-integer-expression test)) nil)
	(t (parse-formula body))))

(defun parse-not (F)
  ;; F is not a literal, that case is handled in parse-formula
  ;; (format t "entering parse-not ~S" F)
  (let ((op (car F)))
    (cond ((eql op 'not) (parse-formula (cadr F)))
          ((eql op 'and) (parse-formula (cons 'or (negate-list (cdr F)))))
          ((eql op 'or) (parse-formula (cons 'and (negate-list (cdr F)))))
          ((eql op 'implies) (parse-formula (list 'and (cadr F) (list 'not (caddr F)))))
	  ((eql op 'if) (cond ((is-true (parse-integer-expression (cadr F))) (parse-formula `(not ,(caddr F))))
			      (t '(()))))
          ((eql op 'equiv) (append (parse-formula `(or ,(cadr F) ,(caddr F)))
                                   (parse-formula `(or (not ,(cadr F)) (not ,(caddr F))))))
          ((eql op 'forall) (parse-formula `(forsome ,(cadr F) ,(caddr F) ,(cadddr F) (not ,(car (cddddr F))))))
          ((eql op 'forsome) (parse-formula `(forall ,(cadr F) ,(caddr F) ,(cadddr F) (not ,(car (cddddr F))))))
          (t (error "Cannot parse negation ~S" F)))))

(defun negate-list (L)
     (cond ((null L) nil)
           (t (cons (list 'not (car L)) (negate-list (cdr L))))))

(defun parse-implies (FL)
  (if (not (= (length FL) 2)) (error "Cannot parse implication ~S" FL))
  (parse-or (cons (list 'not (car FL)) (cdr FL))))

(defun parse-equiv (FL)
  (if (not (= (length FL) 2)) (error "Cannot parse equivalence ~S" FL))
  (append (parse-implies FL)
          (parse-formula `(implies ,(cadr FL) ,(car FL)))))


(defun is-literal (F) 
  (or (is-proposition F)
      (and (eql 'not (car F)) (is-proposition (cadr F)))))

(defun is-proposition (F)
  (or (atom F)
      (not (member (car F) '(and or implies equiv forall forsome not if)))))


(defun parse-and (FL) ;; and just appends the clauses
  (parse-schema-list FL))

(defun parse-or (FL)
  (cond ((null FL) (list nil)) ;; empty OR is the empty clause
	(t (multiply-clauses (parse-schema (car FL)) 
			     (parse-or (cdr FL))))))

(defun multiply-clauses (L R)
  (if (or (null compact-encoding)
	  (< (length L) 2)
	  (< (length R) 2)
	  (< (+ (length L) (length R)) 5))
      (explicit-multiply-clauses L R)
      (let ((g (gensym "XX"))) ;; g selects whether L or R must be true
	(append (mapcar #'(lambda (c) (cons g c)) R) 
		(mapcar #'(lambda (c) (cons (list 'not g) c)) L)))))

(defun merge-clauses (C1 C2)
  (remove-duplicates (append C1 C2) :test #'equal))



(defun explicit-multiply-clauses (L R)
  (cond ((null L) nil)
	(t (append (multiply-one-clause (car L) R)
		   (explicit-multiply-clauses (cdr L) R)))))

(defun multiply-one-clause (C R)
  (cond ((null R) nil)
	(t (append (list (merge-clauses C (car R)))
		   (multiply-one-clause C (cdr R))))))

(defun parse-forall (VAR DOM TEST BODY)
  (cond ((null DOM) nil) ;; the empty list of clauses
        (t (append (parse-binding VAR (car DOM) TEST BODY nil)
                   (parse-forall VAR (cdr DOM) TEST BODY)))))

(defun is-false (x)
  (or (null x) (eql x 0)))

(defun is-true (x)
  (not (is-false x)))

(defun parse-binding (VAR VAL TEST BODY FAILED-TEST-RESULT)
  (declare (special Bind))
  (let ((RESULT FAILED-TEST-RESULT))
    (bind-uniquely VAR VAL)
    (let ((TESTVAL (parse-expression TEST)))
      (if (is-true TESTVAL)
          (setq RESULT (parse-schema BODY))))
    (unbind VAR)
    RESULT))

(defun parse-forsome (VAR DOM TEST BODY)
  (cond ((NULL Dom) (list nil)) ;; the empty clause
        (t (multiply-clauses (parse-binding VAR (car DOM) TEST BODY (list nil))
			     (parse-forsome VAR (cdr DOM) TEST BODY)))))

(defun parse-set-expression (EXPR)
  (let ((answ (parse-expression EXPR)))
    (if (not (listp answ)) (error "Set expected instead of ~S" EXPR))
    answ))

(defun parse-integer-expression (EXPR)
  (let ((answ (parse-expression EXPR)))
    (if (not (integerp answ)) (error "Integer expected instead of ~S" EXPR))
    answ))

(defun parse-entity-expression (EXPR)
  (let ((answ (parse-expression EXPR)))
    (if (or (null answ) (listp answ)) (error "Entity expected instead of ~S" EXPR))
    answ))

(defun parse-or-expression (EXPR)
  (cond ((null EXPR) 0)
	((is-true (parse-integer-expression (car EXPR))) 1)
	(t (parse-or-expression (cdr EXPR)))))

(defun parse-and-expression (EXPR)
  (cond ((null EXPR) 1)
	((is-false (parse-integer-expression (car EXPR))) 0)
	(t (parse-and-expression (cdr EXPR)))))


(defun parse-expression (EXPR)
  (declare (special Bind ObservedPredicates))
  (cond ((and (atom EXPR) (is-bound EXPR)) (binding-of EXPR))
        ((atom EXPR) EXPR)
        (t (let ((op (car EXPR)))
             (cond ((and (atom EXPR) (is-bound EXPR)) (binding-of EXPR))
                   ((atom EXPR) EXPR)
                   ((eql op 'not) (if (is-true (parse-integer-expression (cadr EXPR))) 0 1))
		   ((eql op 'and) (parse-and-expression (cdr EXPR)))
		   ((eql op 'or) (parse-or-expression (cdr EXPR)))
                   ((eql op 'member) (if (member (parse-entity-expression (cadr EXPR)) 
                                                 (parse-set-expression (caddr EXPR))) 
                                         1 0))
                   ((member op '(union intersection set-difference)) 
                    (parse-set-function op (cadr EXPR) (caddr EXPR)))
                   ((member op '(eq neq =)) 
                    (parse-entity-function op (cadr EXPR) (caddr EXPR)))
                   ((member op '(< <= > >= + - * div rem mod ))
                    (parse-integer-function op (cadr EXPR) (caddr EXPR)))
                   ((eql op 'range)
                    (parse-range (parse-integer-expression (cadr EXPR))
                                 (parse-integer-expression (caddr EXPR))))
                   ((eql op 'set)
                    (parse-enumerated-set (cdr EXPR)))
                   ((gethash op ObservedPredicates)
                    (parse-observed-literal-expression EXPR))
                   (t (error "Parser error at ~S" EXPR)))))))

(defun parse-observed-literal-expression (EXPR)
  (declare (special ObservedLiterals))
  (gethash (cons (car EXPR) (map 'list #'parse-entity-expression (cdr EXPR)))
           ObservedLiterals 0))

(defun parse-set-function (OP A1 A2)
  (let ((e1 (parse-set-expression A1))
	(e2 (parse-set-expression A2)))
    (cond ((eql op 'union) (union e1 e2))
	  ((eql op 'intersection) (intersection e1 e2))
	  ((eql op 'set-difference) (set-difference e1 e2))
	  (t (error "Parser error at ~S" OP)))))

(defun parse-enumerated-set (EXPR)
  (cond ((null EXPR) nil)
	(t (cons (parse-entity-expression (car EXPR))
		 (parse-enumerated-set (cdr EXPR))))))

(defun parse-entity-function (OP A1 A2)
  (let ((e1 (parse-entity-expression A1))
	(e2 (parse-entity-expression A2)))
    (cond ((or (eql OP 'eq) (eql op '=)) (if (equalp E1 E2) 1 0))
	  ((eql op 'neq) (if (equalp E1 E2) 0 1))
	  (t (error "Parser error at ~S" OP)))))


(defun parse-integer-function (OP A1 A2)
  (let ((e1 (parse-integer-expression A1))
	(e2 (parse-integer-expression A2)))
    (cond
     ((eql op '<) (if (< e1 e2) 1 0))
     ((eql op '>) (if (> e1 e2) 1 0))
     ((eql op '<=) (if (<= e1 e2) 1 0))
     ((eql op '>=) (if (>= e1 e2) 1 0))
     ((eql op '+) (+ e1 e2))
     ((eql op '-) (- e1 e2))
     ((eql op '*) (* e1 e2))
     ((eql op 'div) (floor (/ e1 e2)))
     ((eql op 'rem) (- e1 (* e2 (floor (/ e1 e2)))))
     ((eql op 'mod) (mod e1 e2))
     (t (error "Cannot parse function ~S" OP)))))

(defun parse-range (LOW HIGH)
  (cond ((> LOW HIGH) nil)
        (t (cons LOW (parse-range (+ 1 LOW) HIGH)))))

(defun parse-literal (LIT)
  (cond ((and (listp LIT) (eq (car LIT) 'not)) 
	 (list 'not (parse-proposition (cadr LIT))))
	(t (parse-proposition LIT))))

(defun parse-proposition (P)
  (cond ((null P) (error "Unexpected empty set"))
	((atom P) P)
	(t (cons (parse-predicate (car P)) (parse-terms (cdr P))))))

(defun parse-predicate (P)
  (declare (special Bind))
  (if (or (null p) (not (atom p)) (is-bound P))
      (error "Expected predicate instead of ~S" P))
  P)

(defun parse-terms (TERMS)
  (cond ((null TERMS) nil)
	(t (cons (parse-term (car TERMS)) (parse-terms (cdr TERMS))))))

(defun parse-term (TERM)
  (cond ((or (atom TERM)
	     (member (car TERM) 
		     '(not + - * div rem mod eq = neq < <= > >= union 
		       intersection set-difference member)))
	 (parse-entity-expression TERM))
	((member TERM '(forall forsome equiv if and or not implies))
	 (error "Unexpected occurence of ~S" TERM))
	(t (list (car TERM) (parse-term (cadr TERM))))))

	     



